# -*- coding: utf-8 -*-

import re


#######################################
#
#   Version Class
#
#######################################


class Version(object):
    """
    Semver representation
    """

    last_number_re = re.compile(r"(?:[^\d]*(\d+)[^\d]*)+")
    semver_re = re.compile(
        r"""
            ^(?P<major>0|[1-9]\d*)
            \.(?P<minor>0|[1-9]\d*)
            \.(?P<patch>0|[1-9]\d*)
            (?:-(?P<prerelease>
                (?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)
                (?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*
            ))?
            (?:\+(?P<build>
                [0-9a-zA-Z-]+
                (?:\.[0-9a-zA-Z-]+)*
            ))?$
        """,
        re.VERBOSE,
    )
    pep440_re = re.compile(
        r'''
        ^([1-9][0-9]*!)?
        (0|[1-9][0-9]*)
        (\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?
        (\.post(0|[1-9][0-9]*))?
        (\.dev(0|[1-9][0-9]*))?$
        ''',
        re.VERBOSE
    )

    def __init__(
        self, 
        major=0, 
        minor=0, 
        patch=0, 
        prerelease=None, 
        build=None):

        self._major = int(major)
        self._minor = int(minor)
        self._patch = int(patch)
        self._prerelease = prerelease
        self._build = build

    @property
    def major(self):
        return self._major

    @major.setter
    def major(self, value):
        raise AttributeError("major is readonly")

    @property
    def minor(self):
        return self._minor

    @minor.setter
    def minor(self, value):
        raise AttributeError("minor is readonly")

    @property
    def patch(self):
        return self._patch

    @patch.setter
    def patch(self, value):
        raise AttributeError("patch is readonly")

    @property
    def prerelease(self):
        return self._prerelease

    @prerelease.setter
    def prerelease(self, value):
        raise AttributeError("prerelease is readonly")

    @property
    def prerelease_as_int(self):
        """
        Parse an int from prerelease... 
        """

        if not self.prerelease:
            return None
        match = Version.last_number_re.search(self.prerelease)
        if match:
            return int(match.group(1))

    @property
    def prerelease_pep440(self):
        """
        Parse an int from prerelease... 
        -beta.5 = 5
        """

        numeric = self.prerelease_as_int
        if numeric is None:
            return None

        if 'alpha' in self.prerelease.lower():
            return f'a{numeric}'
        elif 'beta' in self.prerelease.lower():
            return f'b{numeric}'
        elif 'rc' in self.prerelease.lower():
            return f'rc{numeric}'
        else:
            return f'.dev{numeric}'

    @property
    def build(self):
        return self._build

    @build.setter
    def build(self, value):
        raise AttributeError("build is readonly")

    @property
    def build_as_int(self):
        """
        Parse an int from build... 
        +build.102 = 102
        """

        if not self.build:
            return 0
        match = Version.last_number_re.search(self.build)
        if match:
            return int(match.group(1))

    @property
    def semantic(self):
        """
        This is the standard semantic version format.  For most
        purposes this is the best choice.
        """
        v = f'{self._major}.{self._minor}.{self._patch}'
        if self._prerelease:
            v = f'{v}-{self._prerelease}'
        if self._build:
            v = f'{v}+{self._build}'
        return v

    @property
    def pep440(self):
        """
        This is a loose translation to pep440 compliant.  They 
        allow more than 3 segments of ints, so we can do 
        1.2.3.899b2 where 899 is build # and b2 is prerelease
        """
        v = self.public
        if self.prerelease:
            v = f'{v}{self.prerelease_pep440}'
        if self.build:
            v = f'{v}.{self.build_as_int}'
        return v

    @property
    def is_canonical(self):
        return Version.pep440_re.match(self.pep440) is not None
    
    @property
    def public(self):
        """
        Public Version - Something simple for pypi
        Should be PEP440, but looks like they take plain
        and simple semantic versions.
        """

        return f'{self._major}.{self._minor}.{self._patch}'

    @property
    def git(self):
        """
        This is a git formatted version, made up from metadata
        retrieved with `git describe`.
        """
        raise ValueError('Not implemented yet.')

    def __str__(self):
        return self.semantic

    def bump_major(self):
        self._major = self._major + 1
        self._minor = 0
        self._patch = 0
        self._prerelease = None
        self.bump_build()

    def bump_minor(self):
        self._minor = self._minor + 1
        self._patch = 0
        self._prerelease = None
        self.bump_build()

    def bump_patch(self):
        self._patch = self._patch + 1
        self._prerelease = None
        self.bump_build()

    def bump_prerelease(self):
        """
        Raise the build part of the version
        """

        cls = type(self)
        prerelease = cls._increment_string(self._prerelease or ("beta.0"))
        self._prerelease = prerelease

    def bump_build(self):
        """
        Raise the build part of the version
        """

        cls = type(self)
        build = cls._increment_string(self._build or ("build.0"))
        self._build = build

    @classmethod
    def _increment_string(cls, string):
        """
        Look for the last sequence of number(s) in a string and increment.
        :param string: the string to search for.
        :return: the incremented string
        Source:
        http://code.activestate.com/recipes/442460-increment-numbers-in-a-string/#c1
        """

        match = cls.last_number_re.search(string)
        if match:
            next_ = str(int(match.group(1)) + 1)
            start, end = match.span(1)
            string = string[: max(end - len(next_), start)] + next_ + string[end:]
        return string

    @classmethod
    def parse(cls, version):
        """
        Parse version string to a Version instance.
        """

        match = cls.semver_re.match(version)
        if match is None:
            raise ValueError(f"{version} is not valid semantic version string")

        matched_parts = match.groupdict()
        return cls(**matched_parts)

